#include "ByteClass.hpp"

#include "ConstClassInfo.hpp"

#include <iostream>
#include <list>

ByteClass::ByteClass(const u_int32_t& mgc /*= 0*/,
                     const u_int16_t& min /*= 0*/,
                     const u_int16_t& maj /*= 0*/)
  : magic(mgc),
    minor_version(min),
    major_version(maj),
    constant_pool(),
    access_flags(0),
    this_class(0),
    super_class(0),
    interfaces(),
    fields(),
    methods(),
    attributes() {
}

ByteClass::~ByteClass() {
  for (unsigned int i = 0; i < constant_pool.size(); i++) {
    if (constant_pool[i]) {
      delete constant_pool[i];
      constant_pool[i] = NULL;
    }
  }

  for (unsigned int i = 0; i < fields.size(); i++) {
    if (fields[i]) {
      delete fields[i];
      fields[i] = NULL;
    }
  }

  for (unsigned int i = 0; i < methods.size(); i++) {
    if (methods[i]) {
      delete methods[i];
      methods[i] = NULL;
    }
  }

  for (unsigned int i = 0; i < attributes.size(); i++) {
    if (attributes[i]) {
      delete attributes[i];
      attributes[i] = NULL;
    }
  }
}

const u_int32_t& ByteClass::getMagic() const {
  return magic;
}

const u_int16_t& ByteClass::getMinorVersion() const {
  return minor_version;
}

const u_int16_t& ByteClass::getMajorVersion() const {
  return major_version;
}

const u_int16_t ByteClass::getConstantPoolCount() const {
  return constant_pool.size() + 1;
}

const CPInfo& ByteClass::getCPInfo(const u_int16_t& idx) const {
  if ((idx == 0) || // the 0th entry in the constant_pool is empty
      (idx > constant_pool.size())) { // we have to check for > here
      // as internally, the index is shifted down by one!
    cerr << "Index " << idx << " for constant_pool out of bounds!" << endl;
    throw exception();
  }

  return *(constant_pool[idx-1]);
}

const u_int16_t& ByteClass::getAccessFlags() const {
  return access_flags;
}


const u_int16_t& ByteClass::getThisClass() const {
  return this_class;
}

const u_int16_t& ByteClass::getSuperClass() const {
  return super_class;
}

const u_int16_t ByteClass::getInterfacesCount() const {
  return interfaces.size();
}

const u_int16_t& ByteClass::getInterface(const u_int16_t& idx) const {
  if (idx >= interfaces.size()) {
    cerr << "Index " << idx << " for interfaces out of bounds!" << endl;
    throw exception();
  }

  return interfaces[idx];
}

const u_int16_t ByteClass::getFieldsCount() const {
  return fields.size();
}

const FieldInfo& ByteClass::getFieldInfo(const u_int16_t& idx) const {
  if (idx >= fields.size()) {
    cerr << "Index " << idx << " for fields out of bounds!" << endl;
    throw exception();
  }

  return *(fields[idx]);
}

const u_int16_t ByteClass::getMethodsCount() const {
  return methods.size();
}

const MethodInfo& ByteClass::getMethodInfo(const u_int16_t& idx) const {
  if (idx >= methods.size()) {
    cerr << "Index " << idx << " for methods out of bounds!" << endl;
    throw exception();
  }

  return *(methods[idx]);
}

const u_int16_t ByteClass::getAttributesCount() const {
  return attributes.size();
}

const AttributeInfo& ByteClass::getAttributeInfo(const u_int16_t& idx) const {
  if (idx >= attributes.size()) {
    cerr << "Index " << idx << " for attributes out of bounds!" << endl;
    throw exception();
  }

  return *(attributes[idx]);
}

const JavaString& ByteClass::getThisClassName() const {
  return this_class_name;
}

const JavaString& ByteClass::getSuperClassName() const {
  return super_class_name;
}

list<string> getClassAccessFlags(const u_int16_t& access_flags) {
  list<string> access_flags_list;
  if (access_flags & ByteClass::ACC_PUBLIC) {
    access_flags_list.push_back("public");
  }
  if (access_flags & ByteClass::ACC_FINAL) {
    access_flags_list.push_back("final");
  }
  if (access_flags & ByteClass::ACC_SUPER) {
    access_flags_list.push_back("(resp. invokespecial)");
  }
  if (access_flags & ByteClass::ACC_ABSTRACT) {
    access_flags_list.push_back("abstract");
  }

  return access_flags_list;
}

ostream& ByteClass::toOStreamAsHTML(ostream& os) const {
  os << "<ul>" << endl;
  // magic
  os << "<li>Magic: 0x" << hex << getMagic() << dec << "</li>" << endl;
  // versionb
  os << "<li>Version: " << getMajorVersion() << "." 
     << getMinorVersion() << "</li>" << endl;

  // constant_pool
  os << endl << "<li>ConstantPool (#" << getConstantPoolCount() << "):</li>" << endl;
  os << "<ol>";
  for (u_int16_t i = 1; i < getConstantPoolCount(); i++) {
    if (&getCPInfo(i)) {
      makeReference(os, i, "ConstantPool");
      getCPInfo(i).toOStreamAsHTML(os);
    } else {
      os << "<li/>" << endl;
    }
  }
  os << "</ol>" << endl;

  // access_flags
  os << "<li>AccessFlags: ";
  list<string> afl(getClassAccessFlags(getAccessFlags()));
  for (list<string>::const_iterator af = afl.begin(); af != afl.end(); af++) {
    if (af != afl.begin()) {
      os << ", ";
    }
    os << *af;
  }
  os << "</li>" << endl;

  // this_class
  os << "<li>This Type: ";
  referenceTo(os, getThisClass(), "ConstantPool");
  os << " == " << getThisClassName() << "</li>" << endl;

  // super_class
  u_int16_t suclass(getSuperClass());
  if (suclass > 0) {
    os << "<li>SuperClass:";
    referenceTo(os, suclass, "ConstantPool");
    os << " == " << getSuperClassName() << "</li>" << endl;
  } else {
    os << "<li><b>Type has NO SuperClass, must be java.lang.Object!</b></li>" << endl;
  }

  // interfaces
  os << "<li>Interfaces (#" << getInterfacesCount() << "):</li>" << endl;
  os << "<ul>";
  for (u_int16_t i = 0; i < getInterfacesCount(); i++) {
    os << "<li>";
    referenceTo(os, getInterface(i), "ConstantPool");
    os << " == {" << getCPInfo(getInterface(i))
       << "} == {";
    const ConstClassInfo* const_class_info =
      dynamic_cast<const ConstClassInfo*>(&getCPInfo(getInterface(i)));
    if(const_class_info) {
      os << getCPInfo(const_class_info->getNameIndex());
    } else {
      cerr << "Cannot get ConstClassInfo from Index!" << endl;
      os << "NULL (ParseError)";
    }
    os << "}</li>" << endl;
  }
  os << "</ul>" << endl;

  // fields
  os << "<li>Fields (#" << getFieldsCount() << "):</li>" << endl;
  os << "<ul>";
  for (u_int16_t i = 0; i < getFieldsCount(); i++) {
    getFieldInfo(i).toOStreamAsHTML(os);
  }
  os << "</ul>" << endl;

  // methods
  os << "<li>Methods (#" << getMethodsCount() << "):</li>" << endl;
  os << "<ul>";
  for (u_int16_t i = 0; i < getMethodsCount(); i++) {
    getMethodInfo(i).toOStreamAsHTML(os);
  }
  os << "</ul>" << endl;

  // attributes
  os << "<li>Attributes (#" << getAttributesCount() << "):</li>" << endl;
  os << "<ul>" << endl;
  for (u_int16_t i = 0; i < getAttributesCount(); i++) {
    getAttributeInfo(i).toOStreamAsHTML(os);
  }
  os << "</ul>" << endl;

  os << "</ul>" << endl;

  return os;
}

ostream& ByteClass::makeReference(ostream& os, const u_int16_t& idx, const string& ns) {
  os << "<a name=\"" << ns << idx << "\"/>";

  return os;
}

ostream& ByteClass::referenceTo(ostream& os, const u_int16_t& idx, const string& ns) {
  os << "<a href=\"#" << ns << idx << "\">" << idx << "</a>";

  return os;
}

ostream& operator<<(ostream& os, const ByteClass& byte_class) {
  // magic
  os << "Magic:\t\t0x" << hex << byte_class.getMagic() << dec << endl;
  // version
  os << "Version:\t" << byte_class.getMajorVersion() << "." 
     << byte_class.getMinorVersion() << endl;

  // constant_pool
  os << endl << "ConstantPool (#" << byte_class.getConstantPoolCount() << "):" << endl;
  for (u_int16_t i = 1; i < byte_class.getConstantPoolCount(); i++) {
    if (&byte_class.getCPInfo(i)) {
      os << "#" << i << "\t" << byte_class.getCPInfo(i) << endl;
    }
  }
  os << endl;

  // access_flags
  os << "AccessFlags:\t";
  list<string> afl(getClassAccessFlags(byte_class.getAccessFlags()));
  for (list<string>::const_iterator af = afl.begin(); af != afl.end(); af++) {
    if (af != afl.begin()) {
      os << ", ";
    }
    os << *af;
  }
  os << endl;

  // this_class
  os << "This Type:\t" << byte_class.getThisClass()
     << "\t== " << byte_class.getThisClassName() << endl;

  // super_class
  u_int16_t suclass(byte_class.getSuperClass());
  if (suclass > 0) {
    os << "SuperClass:\t" << suclass
       << "\t== " << byte_class.getSuperClassName() << endl;
  } else {
    os << "Type has NO SuperClass, must be java.lang.Object!" << endl;
  }

  // interfaces
  os << endl << "Interfaces (#" << byte_class.getInterfacesCount() << "):" << endl;
  for (u_int16_t i = 0; i < byte_class.getInterfacesCount(); i++) {
    os << byte_class.getInterface(i) 
       << "\t-> " << byte_class.getCPInfo(byte_class.getInterface(i))
       << "\t-> ";
    const ConstClassInfo* const_class_info =
      dynamic_cast<const ConstClassInfo*>(&byte_class.getCPInfo(byte_class.getInterface(i)));
    if(const_class_info) {
      os << byte_class.getCPInfo(const_class_info->getNameIndex());
    } else {
      cerr << "Cannot get ConstClassInfo from Index!" << endl;
      os << "NULL (ParseError)";
    }
    os << endl;
  }

  // fields
  os << endl << "Fields (#" << byte_class.getFieldsCount() << "):" << endl;
  for (u_int16_t i = 0; i < byte_class.getFieldsCount(); i++) {
    os << byte_class.getFieldInfo(i) << endl;
  }

  // methods
  os << endl << "Methods (#" << byte_class.getMethodsCount() << "):" << endl;
  for (u_int16_t i = 0; i < byte_class.getMethodsCount(); i++) {
    os << byte_class.getMethodInfo(i) << endl;
  }

  // attributes
  os << endl << "Attributes (#" << byte_class.getAttributesCount() << "):" << endl;
  for (u_int16_t i = 0; i < byte_class.getAttributesCount(); i++) {
    os << byte_class.getAttributeInfo(i) << endl;
  }

  return os;
}
