#include "ByteCode.hpp"

#include "ByteClass.hpp"

#include <iomanip>
#include <sstream>

ByteCode::ByteCode(const u_int8_t* c, const u_int32_t& length)
  : code(NULL),
    code_length(length) {
  if (c && (length > 0)) {
    code = new u_int8_t[code_length];
    memcpy(code, c, code_length);
  }
}

ByteCode::~ByteCode() {
  if (code) {
    delete[] code;
    code = NULL;
  }
}

u_int8_t getByteFromCode(const u_int8_t* code,
                         const u_int32_t& code_length,
                         u_int16_t& offset,
                         const string& caller = "ByteCode") {
  if ((unsigned int)offset+1 >= code_length) {
    cerr << "Invalid code: " << caller << " needs 1 bytes argument!" << endl;
    throw exception();
  }
  return code[++offset];
}

int8_t getSignedByteFromCode(const u_int8_t* code,
                             const u_int32_t& code_length,
                             u_int16_t& offset,
                             const string& caller = "ByteCode") {
  if ((unsigned int)offset+1 >= code_length) {
    cerr << "Invalid code: " << caller << " needs 1 bytes argument!" << endl;
    throw exception();
  }
  return code[++offset];
}

u_int16_t getWordFromCode(const u_int8_t* code,
                          const u_int32_t& code_length,
                          u_int16_t& offset,
                          const string& caller = "ByteCode") {
  if ((unsigned int)offset+2 >= code_length) {
    cerr << "Invalid code: " << caller << " needs 2 bytes argument!" << endl;
    throw exception();
  }
  u_int8_t byte1(code[++offset]);
  u_int8_t byte2(code[++offset]);
  return ((((u_int16_t) byte1) << 8) | byte2);
}

int16_t getSignedWordFromCode(const u_int8_t* code,
                              const u_int32_t& code_length,
                              u_int16_t& offset,
                              const string& caller = "ByteCode") {
  if ((unsigned int)offset+2 >= code_length) {
    cerr << "Invalid code: " << caller << " needs 2 bytes argument!" << endl;
    throw exception();
  }
  int8_t byte1(code[++offset]);
  int8_t byte2(code[++offset]);
  return ((((int16_t) byte1) << 8) | byte2);
}

int32_t getSignedDoubleWordFromCode(const u_int8_t* code,
                                    const u_int32_t& code_length,
                                    u_int16_t& offset,
                                    const string& caller = "ByteCode") {
  if ((unsigned int)offset+4 >= code_length) {
    cerr << "Invalid code: " << caller << " needs 4 bytes argument!" << endl;
    throw exception();
  }
  int8_t byte1(code[++offset]);
  int8_t byte2(code[++offset]);
  int8_t byte3(code[++offset]);
  int8_t byte4(code[++offset]);
  return 
    (((int32_t) byte1) << 24)
  | (((int32_t) byte2) << 16)
  | (((int32_t) byte3) << 8 )
  | byte4;
}

ostream& ByteCode::toOStream(ostream& os) const {
  for (u_int16_t i = 0; i < code_length; i++) {
    os.fill('0');
    os.width(3);
    os << i;
    os << " " << hex;
    os.fill('0');
    os.width(2);
    os << (int) code[i] << dec;

    switch ((OPCode) code[i]) {
    // no operatation
    case NOP:
      os << " nop";
      break;

    // pushing constants to stack
    case ACONST_NULL:
      os << " aconst_null";
      break;
    case ICONST_M1:
      os << " iconst_m1";
      break;
    case ICONST_0:
      os << " iconst_0";
      break;
    case ICONST_1:
      os << " iconst_1";
      break;
    case ICONST_2:
      os << " iconst_2";
      break;
    case ICONST_3:
      os << " iconst_3";
      break;
    case ICONST_4:
      os << " iconst_4";
      break;
    case ICONST_5:
      os << " iconst_5";
      break;
    case LCONST_0:
      os << " lconst_0";
      break;
    case LCONST_1:
      os << " lconst_1";
      break;
    case FCONST_0:
      os << " fconst_0";
      break;
    case FCONST_1:
      os << " fconst_1";
      break;
    case FCONST_2:
      os << " fconst_2";
      break;
    case DCONST_0:
      os << " dconst_0";
      break;
    case DCONST_1:
      os << " dconst_1";
      break;
    case BIPUSH:
      os << " bipush " << getSignedByteFromCode(code, code_length, i, "bipush");
      break;
    case SIPUSH:
      os << " sipush " << getSignedWordFromCode(code, code_length, i, "sipush");
      break;

    // pushing constant_pool values to stack
    case LDC:
      os << " ldc idx(";
      ByteClass::referenceTo(os, getByteFromCode(code, code_length, i, "ldc"), "ConstantPool");
      os << ")";
      break;
    case LDC_W:
      os << " ldc_w idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "ldc_w"), "ConstantPool");
      os << ")";
      break;
    case LDC2_W:
      os << " ldc2_w idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "ldc2_w"), "ConstantPool");
      os << ")";
      break;

    // loading from local variables
    case ILOAD:
      os << " iload " << (int) getByteFromCode(code, code_length, i, "iload");
      break;
    case LLOAD:
      os << " lload " << (int) getByteFromCode(code, code_length, i, "lload");
      break;
    case FLOAD:
      os << " fload " << (int) getByteFromCode(code, code_length, i, "fload");
      break;
    case DLOAD:
      os << " dload " << (int) getByteFromCode(code, code_length, i, "dload");
      break;
    case ALOAD:
      os << " aload " << (int) getByteFromCode(code, code_length, i, "aload");
      break;
    case ILOAD_0:
      os << " iload_0";
      break;
    case ILOAD_1:
      os << " iload_1";
      break;
    case ILOAD_2:
      os << " iload_2";
      break;
    case ILOAD_3:
      os << " iload_3";
      break;
    case LLOAD_0:
      os << " lload_0";
      break;
    case LLOAD_1:
      os << " lload_1";
      break;
    case LLOAD_2:
      os << " lload_2";
      break;
    case LLOAD_3:
      os << " lload_3";
      break;
    case FLOAD_0:
      os << " fload_0";
      break;
    case FLOAD_1:
      os << " fload_1";
      break;
    case FLOAD_2:
      os << " fload_2";
      break;
    case FLOAD_3:
      os << " fload_3";
      break;
    case DLOAD_0:
      os << " dload_0";
      break;
    case DLOAD_1:
      os << " dload_1";
      break;
    case DLOAD_2:
      os << " dload_2";
      break;
    case DLOAD_3:
      os << " dload_3";
      break;
    case ALOAD_0:
      os << " aload_0";
      break;
    case ALOAD_1:
      os << " aload_1";
      break;
    case ALOAD_2:
      os << " aload_2";
      break;
    case ALOAD_3:
      os << " aload_3";
      break;
    case IALOAD:
      os << " iaload";
      break;
    case LALOAD:
      os << " laload";
      break;
    case FALOAD:
      os << " faload";
      break;
    case DALOAD:
      os << " daload";
      break;
    case AALOAD:
      os << " aaload";
      break;
    case BALOAD:
      os << " baload";
      break;
    case CALOAD:
      os << " caload";
      break;
    case SALOAD:
      os << " saload";
      break;

    // storing to local variables
    case ISTORE:
      os << " istore " << (int) getByteFromCode(code, code_length, i, "istore");
      break;
    case LSTORE:
      os << " lstore " << (int) getByteFromCode(code, code_length, i, "lstore");
      break;
    case FSTORE:
      os << " fstore " << (int) getByteFromCode(code, code_length, i, "fstore");
      break;
    case DSTORE:
      os << " dstore " << (int) getByteFromCode(code, code_length, i, "dstore");
      break;
    case ASTORE:
      os << " astore " << (int) getByteFromCode(code, code_length, i, "astore");
      break;
    case ISTORE_0:
      os << " istore_0";
      break;
    case ISTORE_1:
      os << " istore_1";
      break;
    case ISTORE_2:
      os << " istore_2";
      break;
    case ISTORE_3:
      os << " istore_3";
      break;
    case LSTORE_0:
      os << " lstore_0";
      break;
    case LSTORE_1:
      os << " lstore_1";
      break;
    case LSTORE_2:
      os << " lstore_2";
      break;
    case LSTORE_3:
      os << " lstore_3";
      break;
    case FSTORE_0:
      os << " fstore_0";
      break;
    case FSTORE_1:
      os << " fstore_1";
      break;
    case FSTORE_2:
      os << " fstore_2";
      break;
    case FSTORE_3:
      os << " fstore_3";
      break;
    case DSTORE_0:
      os << " dstore_0";
      break;
    case DSTORE_1:
      os << " dstore_1";
      break;
    case DSTORE_2:
      os << " dstore_2";
      break;
    case DSTORE_3:
      os << " dstore_3";
      break;
    case ASTORE_0:
      os << " astore_0";
      break;
    case ASTORE_1:
      os << " astore_1";
      break;
    case ASTORE_2:
      os << " astore_2";
      break;
    case ASTORE_3:
      os << " astore_3";
      break;
    case IASTORE:
      os << " iastore";
      break;
    case LASTORE:
      os << " lastore";
      break;
    case FASTORE:
      os << " fastore";
      break;
    case DASTORE:
      os << " dastore";
      break;
    case AASTORE:
      os << " aastore";
      break;
    case BASTORE:
      os << " bastore";
      break;
    case CASTORE:
      os << " castore";
      break;
    case SASTORE:
      os << " sastore";
      break;

    // generic stack operations
    case POP:
      os << " pop";
      break;
    case POP2:
      os << " pop2";
      break;
    case DUP:
      os << " dup";
      break;
    case DUP_X1:
      os << " dup_x1";
      break;
    case DUP_x2:
      os << " dup_x2";
      break;
    case DUP2:
      os << " dup2";
      break;
    case DUP2_X1:
      os << " dup2_x1";
      break;
    case DUP2_X2:
      os << " dup2_x2";
      break;
    case SWAP:
      os << " swap";
      break;

    // integer and float aritmetic
    case IADD:
      os << " iadd";
      break;
    case LADD:
      os << " ladd";
      break;
    case FADD:
      os << " fadd";
      break;
    case DADD:
      os << " dadd";
      break;
    case ISUB:
      os << " isub";
      break;
    case LSUB:
      os << " lsub";
      break;
    case DSUB:
      os << " dsub";
      break;
    case FSUB:
      os << " fsub";
      break;
    case IMUL:
      os << " imul";
      break;
    case LMUL:
      os << " lmul";
      break;
    case FMUL:
      os << " fmul";
      break;
    case DMUL:
      os << " dmul";
      break;
    case IDIV:
      os << " idiv";
      break;
    case LDIV:
      os << " ldiv";
      break;
    case FDIV:
      os << " fdiv";
      break;
    case DDIV:
      os << " ddiv";
      break;
    case IREM:
      os << " irem";
      break;
    case LREM:
      os << " lrem";
      break;
    case FREM:
      os << " frem";
      break;
    case DREM:
      os << " drem";
      break;
    case INEG:
      os << " ineg";
      break;
    case LNEG:
      os << " lneg";
      break;
    case FNEG:
      os << " fneg";
      break;
    case DNEG:
      os << " dneg";
      break;
    case IINC:
      os << " iinc " << (int) getByteFromCode(code, code_length, i, "iinc")
         << " " << (int) getByteFromCode(code, code_length, i, "iinc") << "";
      break;

    // logic
    case ISHL:
      os << " ishl";
      break;
    case LSHL:
      os << " lshl";
      break;
    case ISHR:
      os << " ishr";
      break;
    case LSHR:
      os << " lshr";
      break;
    case IUSHR:
      os << " iushr";
      break;
    case LUSHR:
      os << " lushr";
      break;
    case IAND:
      os << " iand";
      break;
    case LAND:
      os << " land";
      break;
    case IOR:
      os << " ior";
      break;
    case LOR:
      os << " lor";
      break;
    case IXOR:
      os << " ixor";
      break;
    case LXOR:
      os << " lxor";
      break;

    // type conversion
    case I2L:
      os << " i2l";
      break;
    case I2F:
      os << " i2f";
      break;
    case I2D:
      os << " i2d";
      break;
    case L2L:
      os << " l2l";
      break;
    case L2F:
      os << " l2f";
      break;
    case L2D:
      os << " l2d";
      break;
    case F2I:
      os << " f2i";
      break;
    case F2L:
      os << " f2l";
      break;
    case F2D:
      os << " f2d";
      break;
    case D2I:
      os << " d2i";
      break;
    case D2L:
      os << " d2l";
      break;
    case D2F:
      os << " d2f";
      break;
    case I2B:
      os << " i2b";
      break;
    case I2C:
      os << " i2c";
      break;
    case I2S:
      os << " i2s";
      break;

    // control flow
    case LCMP:
      os << " lcmp";
      break;
    case FCMPL:
      os << " fcmpl";
      break;
    case FCMPG:
      os << " fcmpg";
      break;
    case DCMPL:
      os << " dcmpl";
      break;
    case DCMPG:
      os << " dcmpg";
      break;

    case IFEQ:
      os << " ifeq ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifeq");
      break;
    case IFNE:
      os << " ifne ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifne");
      break;
    case IFLT:
      os << " iflt ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "iflt");
      break;
    case IFGE:
      os << " ifge ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifge");
      break;
    case IFGT:
      os << " ifgt ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifgt");
      break;
    case IFLE:
      os << " ifle ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifle");
      break;

    case IF_ICMPEQ:
      os << " if_icmpeq ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_icmpeq");
      break;
    case IF_ICMPNE:
      os << " if_icmpne ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_icmpne");
      break;
    case IF_ICMPLT:
      os << " if_icmplt ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_icmplt");
      break;
    case IF_ICMPGE:
      os << " if_icmpge ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_icmpge");
      break;
    case IF_ICMPGT:
      os << " if_icmpgt ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_icmpgt");
      break;
    case IF_ICMPLE:
      os << " if_icmple ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_icmple");
      break;
    case IF_ACMPEQ:
      os << " if_acmpeq ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_acmpeq");
      break;
    case IF_ACMPNE:
      os << " if_acmpne ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "if_acmpne");
      break;
    case GOTO:
      os << " goto ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "goto");
      break;

    case JSR:
      os << " jsr ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "jsr");
      break;
    case RET:
      os << " ret " << (int) getByteFromCode(code, code_length, i, "ret");
      break;

    case TABLESWITCH:
      os << " tableswitch";
      {
        // save the bytecode position of the loopupswitch opcode
        u_int16_t current_pc(i);
        // search for 32Bit alignment
        while ( ((i+1) % 4) != 0) {
          i++;
          if (i >= code_length) {
            cerr << "Cannot skip alignment bytes for tableswitch instruction!" << endl;
            throw exception();
          }
        }
        // save the default value later printing
        int32_t defaultvalue(current_pc + getSignedDoubleWordFromCode(code, code_length, i, "tableswitch"));
        int32_t low(getSignedDoubleWordFromCode(code, code_length, i, "tableswitch"));
        int32_t high(getSignedDoubleWordFromCode(code, code_length, i, "tableswitch"));
        os << " from " << low << " to " << high << " default: " << defaultvalue;
        // read and print the offsets
        for (int32_t idx = low; idx <= high; idx++) {
          int32_t offset(current_pc + getSignedDoubleWordFromCode(code, code_length, i, "tableswitch"));
          os << endl
             << "          " << idx << ":\t" << offset;
        }
      }
      break;

    case LOOKUPSWITCH:
      os << " lookupswitch";
      {
        // save the bytecode position of the loopupswitch opcode
        u_int16_t current_pc(i);
        // search for 32Bit alignment
        while ( ((i+1) % 4) != 0) {
          i++;
          if (i >= code_length) {
            cerr << "Cannot skip alignment bytes for lookupswitch instruction!" << endl;
            throw exception();
          }
        }
        // save the default value for printing at last
        int32_t defaultvalue(current_pc + getSignedDoubleWordFromCode(code, code_length, i, "lookupswitch"));
        // how many match-offset pairs do we have
        int32_t npairs(getSignedDoubleWordFromCode(code, code_length, i, "lookupswitch"));
        // read and print the match-offset pairs
        for (int32_t pair = 0; pair < npairs; pair++) {
          int32_t match(getSignedDoubleWordFromCode(code, code_length, i, "lookupswitch"));
          int32_t offset(current_pc + getSignedDoubleWordFromCode(code, code_length, i, "lookupswitch"));
          os << endl
             << "          " << match << ":\t" << offset;
        }
        // print the default value previously stored
        os << endl
           <<   "          default:\t" << defaultvalue;
      }
      break;

    case IFNULL:
      os << " ifnull ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifnull");
      break;
    case IFNONNULL:
      os << " ifnonnull ";
      os.width(3);
      os.fill('0');
      os << i + getSignedWordFromCode(code, code_length, i, "ifnonnull");
      break;
    case GOTO_W:
      os << " goto_w ";
      os.width(3);
      os.fill('0');
      os << i + getSignedDoubleWordFromCode(code, code_length, i, "goto_w");
      break;
    case JSR_W:
      os << " jsr_w ";
      os.width(3);
      os.fill('0');
      os << i + getSignedDoubleWordFromCode(code, code_length, i, "jsr_w");
      break;

    // method invocation and return
    case IRETURN:
      os << " ireturn";
      break;
    case LRETURN:
      os << " lreturn";
      break;
    case FRETURN:
      os << " freturn";
      break;
    case DRETURN:
      os << " dreturn";
      break;
    case ARETURN:
      os << " areturn";
      break;
    case RETURN:
      os << " return";
      break;
    case INVOKEVIRTUAL:
      os << " invokevirtual idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "invokevirtual"), "ConstantPool");
      os << ")";
      break;
    case INVOKESPECIAL:
      os << " invokespecial idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "invokespecial"), "ConstantPool");
      os << ")";
      break;
    case INVOKESTATIC:
      os << " invokestatic idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "invokestatic"), "ConstantPool");
      os << ")";
      break;
    case INVOKEINTERFACE:
      os << " invokeinterface idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "invokeinterface"), "ConstantPool");
      os << ") count(";
      ByteClass::referenceTo(os, getByteFromCode(code, code_length, i, "invokeinterface"), "ConstantPool");
      os << ")";
      // SPEC: "The value of the fourth operand must always be 0."
      // We simple ignore it!
      getWordFromCode(code, code_length, i, "invokeinterface");
      break;

    // objects and arrays
    case GETSTATIC:
      os << " getstatic idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "getstatic"), "ConstantPool");
      os << ")";
      break;
    case PUTSTATIC:
      os << " putstatic idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "putstatic"), "ConstantPool");
      os << ")";
      break;
    case GETFIELD:
      os << " getfield idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "getfield"), "ConstantPool");
      os << ")";
      break;
      break;
    case PUTFIELD:
      os << " putfield idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "putfield"), "ConstantPool");
      os << ")";
      break;

    case NEW:
      os << " new idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "new"), "ConstantPool");
      os << ")";
      break;
    case NEWARRAY:
      os << " newarray ";
      switch (getByteFromCode(code, code_length, i, "newarray")) {
      case 4:
        os << "T_BOOLEAN";
        break;
      case 5:
        os << "T_CHAR";
        break;
      case 6:
        os << "T_FLOAT";
        break;
      case 7:
        os << "T_DOUBLE";
        break;
      case 8:
        os << "T_BYTE";
        break;
      case 9:
        os << "T_SHORT";
        break;
      case 10:
        os << "T_INT";
        break;
      case 11:
        os << "T_LONG";
        break;
      default:
        cerr << "Unknown array type in newarray!" << endl;
        throw exception();
        break;
      }
      break;
    case ANEWARRAY:
      os << " anewarray idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "anewarray"), "ConstantPool");
      os << ")";
      break;
    case ARRAYLENGTH:
      os << " arraylength";
      break;
    case CHECKCAST:
      os << " checkcast idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "checkcast"), "ConstantPool");
      os << ")";
      break;
    case INSTANCEOF:
      os << " instanceof idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "instanceof"), "ConstantPool");
      os << ")";
      break;
    case MULTIANEWARRAY:
      os << " multianewarray idx(";
      ByteClass::referenceTo(os, getWordFromCode(code, code_length, i, "multianewarray"), "ConstantPool");
      os << ") dim(" << (int) getByteFromCode(code, code_length, i, "multianewarray") << ")";
      break;

    // exceptions
    case ATHROW:
      os << " athrow";
      break;

    // thread synchronisation
    case MONITORENTER:
      os << " monitorenter";
      break;
    case MONITOREXIT:
      os << " monitorexit";
      break;

    // the wide instruction
    case WIDE:
      os << " wide ";
      OPCode instruction((OPCode) getByteFromCode(code, code_length, i, "wide"));
      u_int16_t index(getWordFromCode(code, code_length, i, "wide"));
      switch (instruction) {
      case ILOAD:
        os << " iload " << index;
        break;
      case FLOAD:
        os << " fload " << index;
        break;
      case ALOAD:
        os << " aload " << index;
        break;
      case LLOAD:
        os << " lload " << index;
        break;
      case DLOAD:
        os << " dload " << index;
        break;

      case ISTORE:
        os << " istore " << index;
        break;
      case FSTORE:
        os << " fstore " << index;
        break;
      case ASTORE:
        os << " astore " << index;
        break;
      case LSTORE:
        os << " lstore " << index;
        break;
      case DSTORE:
        os << " dstore " << index;
        break;

      case RET:
        os << " ret " << index;
        break;

      case IINC:
        os << " iinc " << index << " " << getSignedWordFromCode(code, code_length, i, "wide iinc");
        break;

      default:
        cerr << "Illegal opcode for wide instruction application: " << instruction << endl;
        throw exception();
      }
      break;
    }

    os << endl;
  }

  return os;
}

ostream& operator<<(ostream& os, const ByteCode& bc) {
  return bc.toOStream(os);
}
