#include "intconvert.h"

#include "ClassLoader.hpp"

#include "ConstUtf8Info.hpp"
#include "ConstClassInfo.hpp"
#include "ConstStringInfo.hpp"
#include "ConstIntegerInfo.hpp"
#include "ConstLongInfo.hpp"
#include "ConstFieldInfo.hpp"
#include "ConstMethodInfo.hpp"
#include "ConstInterfaceMethodInfo.hpp"
#include "ConstNameAndTypeInfo.hpp"
#include "ConstFloatInfo.hpp"
#include "ConstDoubleInfo.hpp"

#include "SourceFileAttribute.hpp"
#include "CodeAttribute.hpp"
#include "LineNumberTableAttribute.hpp"
#include "ConstantValueAttribute.hpp"
#include "ExceptionsAttribute.hpp"
#include "InnerClassesAttribute.hpp"
#include "SyntheticAttribute.hpp"
#include "LocalVariableTableAttribute.hpp"
#include "DeprecatedAttribute.hpp"

#include <iostream>

ClassLoader::ClassLoader()
  : filename(""),
    input() {
}

ClassLoader::~ClassLoader() {
  if (input.is_open()) {
    cerr << "Closing input in D'tor!" << endl;
    input.close();
  }
}

void ClassLoader::read(u_int8_t& val) {
  if (!input.is_open()) {
    cerr << "No file opened!" << endl;
    throw exception();
  }

  input.read((char*) &val, sizeof(u_int8_t));
}

void ClassLoader::read(u_int16_t& val) {
  if (!input.is_open()) {
    cerr << "No file opened!" << endl;
    throw exception();
  }

  u_int16_t buf;
  input.read((char*) &buf, sizeof(u_int16_t));

  val = be16_to_cpu(buf);
}

void ClassLoader::read(u_int32_t& val) {
  if (!input.is_open()) {
    cerr << "No file opened!" << endl;
    throw exception();
  }

  u_int32_t buf;
  input.read((char*) &buf, sizeof(u_int32_t));

  val = be32_to_cpu(buf);
}

void ClassLoader::read(u_int64_t& val) {
  if (!input.is_open()) {
    cerr << "No file opened!" << endl;
    throw exception();
  }

  u_int64_t buf;
  input.read((char*) &buf, sizeof(u_int64_t));

  val = be64_to_cpu(buf);
}


void ClassLoader::open() {
  input.open(filename.c_str());

  if (input.fail()) {
    throw exception();
  }
}

void ClassLoader::checkHeader() {
}

CPInfo* ClassLoader::parseCPInfo(bool& skip) {
  skip = false;

  // first we read the TAG
  u_int8_t tag;
  read(tag);

  switch (tag) {
  case CPInfo::CONSTANT_Utf8:
    {
      // we need the length of the Utf8 data from the bytecode
      u_int16_t length(0);
      read(length);
      u_int8_t* info = new u_int8_t[length];
      input.read((char*) info, length);
      return new ConstUtf8Info(length, info);
    }
    break;

  case CPInfo::CONSTANT_Class:
    {
      u_int16_t name_index(0);
      read(name_index);
      return new ConstClassInfo(name_index);
    }
    break;

  case CPInfo::CONSTANT_String:
    {
      u_int16_t string_index(0);
      read(string_index);
      return new ConstStringInfo(string_index);
    }
    break;

  case CPInfo::CONSTANT_Integer:
    {
      u_int32_t integer(0);
      read(integer);
      return new ConstIntegerInfo(integer);
    }
    break;

  case CPInfo::CONSTANT_Fieldref:
    {
      u_int16_t class_index(0);
      read(class_index);
      u_int16_t name_and_type_index(0);
      read(name_and_type_index);
      return new ConstFieldInfo(class_index, name_and_type_index);
    }
    break;

  case CPInfo::CONSTANT_Methodref:
    {
      u_int16_t class_index(0);
      read(class_index);
      u_int16_t name_and_type_index(0);
      read(name_and_type_index);
      return new ConstMethodInfo(class_index, name_and_type_index);
    }
    break;

  case CPInfo::CONSTANT_InterfaceMethodref:
    {
      u_int16_t class_index(0);
      read(class_index);
      u_int16_t name_and_type_index(0);
      read(name_and_type_index);
      return new ConstInterfaceMethodInfo(class_index, name_and_type_index);
    }
    break;

  case CPInfo::CONSTANT_NameAndType:
    {
      u_int16_t name_index(0);
      read(name_index);
      u_int16_t descriptor_index(0);
      read(descriptor_index);
      return new ConstNameAndTypeInfo(name_index, descriptor_index);
    }
    break;

  case CPInfo::CONSTANT_Long:
    {
      // from JVM Spec:
      // "In retrospect, making 8-byte constants take TWO constant
      // pool entries was a poor choise."
      skip = true;
      u_int64_t long_integer(0);
      read(long_integer);
      return new ConstLongInfo(long_integer);
    }
    break;

  case CPInfo::CONSTANT_Float:
    {
      u_int32_t float_data(0);
      read(float_data);
      return new ConstFloatInfo(float_data);
    }
    break;

  case CPInfo::CONSTANT_Double:
    {
      // from JVM Spec:
      // "In retrospect, making 8-byte constants take TWO constant
      // pool entries was a poor choise."
      skip = true;
      u_int64_t double_data(0);
      read(double_data);
      return new ConstDoubleInfo(double_data);
    }
    break;

  default:
    cerr << "Unkown ConstantPool entry type!" << endl;
    throw exception();
    break;
  }

  // if we are here, the C++-compiler is for winblows :-)
  return NULL;
}

AttributeInfo* ClassLoader::parseAttributeInfo(const vector<CPInfo*>& constant_pool) {
  u_int16_t attribute_name_index;
  read(attribute_name_index);

  u_int32_t attribute_length;
  read(attribute_length);
  
  string attribute_name(dynamic_cast<const ConstUtf8Info*>(constant_pool[attribute_name_index-1])->getString().toSTLString());
  if (attribute_name == AttributeInfo::SourceFile) {
    u_int16_t sourcefile_index;
    read(sourcefile_index);
    return new SourceFileAttribute(attribute_name_index, sourcefile_index);
  }

  else if (attribute_name == AttributeInfo::Code) {
    u_int16_t max_stack;
    read(max_stack);
    u_int16_t max_locals;
    read(max_locals);
    u_int32_t code_length;
    read(code_length);
    u_int8_t code[code_length];
    input.read((char*) code, code_length);
    CodeAttribute* code_attribute = 
      new CodeAttribute(attribute_name_index, max_stack, max_locals, code_length, code);
    u_int16_t exception_table_length;
    read(exception_table_length);
    code_attribute->exception_table =
      vector<CodeAttribute::ExceptionTable*>(exception_table_length);
    for (unsigned int i = 0; i < exception_table_length; i++) {
      u_int16_t start_pc;
      read(start_pc);
      u_int16_t end_pc;
      read(end_pc);
      u_int16_t handler_pc;
      read(handler_pc);
      u_int16_t catch_type;
      read(catch_type);
      code_attribute->exception_table[i] =
        new CodeAttribute::ExceptionTable(start_pc, end_pc, handler_pc, catch_type);
    }
    u_int16_t attributes_count;
    read(attributes_count);
    code_attribute->attributes = vector<AttributeInfo*>(attributes_count);
    for (unsigned int i = 0; i < attributes_count; i++) {
      code_attribute->attributes[i] = parseAttributeInfo(constant_pool);
    }      
    return code_attribute;
  }

  else if (attribute_name == AttributeInfo::LineNumberTable) {
    LineNumberTableAttribute* line_number_table_attribute =
      new LineNumberTableAttribute(attribute_name_index);
    u_int16_t line_number_table_length;
    read(line_number_table_length);
    line_number_table_attribute->line_number_table = 
      vector<LineNumberTableAttribute::LineNumberTable*>(line_number_table_length);
    for (unsigned int i = 0; i < line_number_table_length; i++) {
      u_int16_t start_pc;
      read(start_pc);
      u_int16_t line_number;
      read(line_number);
      line_number_table_attribute->line_number_table[i] =
        new LineNumberTableAttribute::LineNumberTable(start_pc, line_number);
    }
    return line_number_table_attribute;
  }

  else if (attribute_name == AttributeInfo::ConstantValue) {
    u_int16_t constantvalue_index;
    read(constantvalue_index);
    return new ConstantValueAttribute(attribute_name_index, constantvalue_index);
  }

  else if (attribute_name == AttributeInfo::Exceptions) {
    ExceptionsAttribute* exceptions_attribute =
      new ExceptionsAttribute(attribute_name_index);
    u_int16_t number_of_exceptions;
    read(number_of_exceptions);
    exceptions_attribute->exception_index_table =
      vector<u_int16_t>(number_of_exceptions);
    for (unsigned int i = 0; i < number_of_exceptions; i++) {
      u_int16_t exception_index;
      read(exception_index);
      exceptions_attribute->exception_index_table[i] = exception_index;
    }
    return exceptions_attribute;
  }

  else if (attribute_name == AttributeInfo::InnerClasses) {
    InnerClassesAttribute* inner_classes_attribute =
      new InnerClassesAttribute(attribute_name_index);
    u_int16_t number_of_inner_classes;
    read(number_of_inner_classes);
    inner_classes_attribute->inner_classes_index_table =
      vector<u_int16_t>(number_of_inner_classes);
    for (unsigned int i = 0; i < number_of_inner_classes; i++) {
      u_int16_t inner_classes_index;
      read(inner_classes_index);
      inner_classes_attribute->inner_classes_index_table[i] = inner_classes_index;
    }
    return inner_classes_attribute;
  }

  else if (attribute_name == AttributeInfo::Synthetic) {
    SyntheticAttribute* synthetic_attribute =
      new SyntheticAttribute(attribute_name_index);
    return synthetic_attribute;
  }

  else if (attribute_name == AttributeInfo::LocalVariableTable) {
    LocalVariableTableAttribute* local_variable_table_attribute =
      new LocalVariableTableAttribute(attribute_name_index);
    u_int16_t local_variable_table_length;
    read(local_variable_table_length);
    local_variable_table_attribute->local_variable_table = 
      vector<LocalVariableTableAttribute::LocalVariableTable*>(local_variable_table_length);
    for (unsigned int i = 0; i < local_variable_table_length; i++) {
      u_int16_t start_pc;
      read(start_pc);
      u_int16_t length;
      read(length);
      u_int16_t name_index;
      read(name_index);
      u_int16_t descriptor_index;
      read(descriptor_index);
      u_int16_t index;
      read(index);
      local_variable_table_attribute->local_variable_table[i] =
        new LocalVariableTableAttribute::LocalVariableTable(start_pc,
                                                            length,
                                                            name_index,
                                                            descriptor_index,
                                                            index);
    }
    return local_variable_table_attribute;
  }

  else if (attribute_name == AttributeInfo::Deprecated) {
    DeprecatedAttribute* deprecated_attribute =
      new DeprecatedAttribute(attribute_name_index);
    return deprecated_attribute;
  }

  u_int8_t* info = NULL;
  if (attribute_length > 0) {
    info = new u_int8_t[attribute_length];
    input.read((char*) info, attribute_length);
  }
    
  return new AttributeInfo(attribute_name_index, attribute_length, info);
}

FieldInfo* ClassLoader::parseFieldInfo(const vector<CPInfo*>& constant_pool) {
  // first the acces flags
  u_int16_t access_flags;
  read(access_flags);

  // now the name_index
  u_int16_t name_index;
  read(name_index);

  // after that the descriptor index to the constant_pool
  u_int16_t descriptor_index;
  read(descriptor_index);

  // how many attributes has this field
  u_int16_t attributes_count;
  read(attributes_count);

  // build the entry
  FieldInfo* field_info = new FieldInfo(access_flags, name_index, descriptor_index, attributes_count);

  // and fill the attributes vector with the parsed values
  for (u_int16_t fattr = 0; fattr < attributes_count; fattr++) {
    field_info->attributes[fattr] = parseAttributeInfo(constant_pool);
  }

  // now return it
  return field_info;
}

MethodInfo* ClassLoader::parseMethodInfo(const vector<CPInfo*>& constant_pool) {
  // first the acces flags
  u_int16_t access_flags;
  read(access_flags);

  // now the name_index
  u_int16_t name_index;
  read(name_index);

  // after that the descriptor index to the constant_pool
  u_int16_t descriptor_index;
  read(descriptor_index);

  // how many attributes has this field
  u_int16_t attributes_count;
  read(attributes_count);

  // build the entry
  MethodInfo* method_info = new MethodInfo(access_flags, name_index, descriptor_index, attributes_count);

  // and fill the attributes vector with the parsed values
  for (u_int16_t mattr = 0; mattr < attributes_count; mattr++) {
    method_info->attributes[mattr] = parseAttributeInfo(constant_pool);
  }

  // now return it
  return method_info;
//   return (MethodInfo*) parseFieldInfo();
}

ByteClass* ClassLoader::load(const string& path) {
  filename = path;

  open();

  // check for 0xCAFEBABE header
  checkHeader();
  u_int32_t magic;
  read(magic);

  if (magic != 0xCAFEBABE) {
    cerr << "Not a Java class file: " << filename << endl;
    throw exception();
  }

  // read version numbers
  u_int16_t minor;
  read(minor);
  u_int16_t major;
  read(major);

  ByteClass* byte_class = new ByteClass(magic, minor, major);

  try {
    // read constant pool length
    u_int16_t constant_pool_count(0);
    read(constant_pool_count);

    // read in constant pool entries
    byte_class->constant_pool = vector<CPInfo*>(constant_pool_count-1);
    for (u_int16_t cpentry = 0; cpentry < constant_pool_count-1; cpentry++) {
      bool skip(false);
      byte_class->constant_pool[cpentry] = parseCPInfo(skip);
      if (skip) {
        cpentry++;
        byte_class->constant_pool[cpentry] = NULL;
      }
    }

    // read the access flags field
    read(byte_class->access_flags);

    // read the this_class index to the constant_pool
    read(byte_class->this_class);
    const CPInfo& this_class_cpinfo(byte_class->getCPInfo(byte_class->getThisClass()));
    const ConstClassInfo& this_class_info(dynamic_cast<const ConstClassInfo&>(this_class_cpinfo));
    const CPInfo& this_class_name_cpinfo(byte_class->getCPInfo(this_class_info.getNameIndex()));
    const ConstUtf8Info& this_class_name_utf8info(dynamic_cast<const ConstUtf8Info&>(this_class_name_cpinfo));
    byte_class->this_class_name = this_class_name_utf8info.getString();

    // read the super_class index to the constant_pool
    read(byte_class->super_class);
    const CPInfo& super_class_cpinfo(byte_class->getCPInfo(byte_class->getSuperClass()));
    const ConstClassInfo& super_class_info(dynamic_cast<const ConstClassInfo&>(super_class_cpinfo));
    const CPInfo& super_class_name_cpinfo(byte_class->getCPInfo(super_class_info.getNameIndex()));
    const ConstUtf8Info& super_class_name_utf8info(dynamic_cast<const ConstUtf8Info&>(super_class_name_cpinfo));
    byte_class->super_class_name = super_class_name_utf8info.getString();

    // how many intarfaces do we inherit
    u_int16_t interfaces_count;
    read(interfaces_count);

    // parse the interfaces
    byte_class->interfaces = vector<u_int16_t>(interfaces_count);
    for (u_int16_t iface = 0; iface < interfaces_count; iface++) {
      read(byte_class->interfaces[iface]);
    }

    // how many fields do we have
    u_int16_t fields_count;
    read(fields_count);

    // parse the field_info table
    byte_class->fields = vector<FieldInfo*>(fields_count);
    for (u_int16_t fd = 0; fd < fields_count; fd++) {
      byte_class->fields[fd] = parseFieldInfo(byte_class->constant_pool);
    }

    // how many methods do we have in this class
    u_int16_t methods_count;
    read(methods_count);

    // parse the method_info table
    byte_class->methods = vector<MethodInfo*>(methods_count);
    for (u_int16_t m = 0; m < methods_count; m++) {
      byte_class->methods[m] = parseMethodInfo(byte_class->constant_pool);
    }

    // how many attributes has this class
    u_int16_t attributes_count;
    read(attributes_count);

    // parse the attributes
    byte_class->attributes = vector<AttributeInfo*>(attributes_count);
    for (u_int16_t attr = 0; attr < attributes_count; attr++) {
      byte_class->attributes[attr] = parseAttributeInfo(byte_class->constant_pool);
    }
  }
  catch (...) {
    cerr << "Error while parsing Bytecode!" << endl;
    delete byte_class;
    input.close();
    throw;
  }

  input.close();

  return byte_class;
}
