#include "CodeAttribute.hpp"

#include "ByteCode.hpp"

CodeAttribute::CodeAttribute(const u_int16_t& index /*= 0*/,
                             const u_int16_t& mstack /*= 0*/,
                             const u_int16_t& mlocals /*= 0*/,
                             const u_int32_t& clength /*= 0*/,
                             const u_int8_t* c /*= NULL*/)
  : AttributeInfo(index, 0, NULL),
    max_stack(mstack),
    max_locals(mlocals),
    code_length(clength),
    code(NULL),
    exception_table(),
    attributes() {
  code = new u_int8_t[code_length];
  memcpy(code, c, code_length);
}

CodeAttribute::~CodeAttribute() {
  if (code) {
    delete[] code;
    code = NULL;
  }
}

const u_int16_t& CodeAttribute::getMaxStack() const {
  return max_stack;
}

const u_int16_t& CodeAttribute::getMaxLocals() const {
  return max_locals;
}

const u_int32_t& CodeAttribute::getCodeLength() const {
  return code_length;
}

u_int16_t CodeAttribute::getExceptionTableLength() const {
  return exception_table.size();
}

const CodeAttribute::ExceptionTable& CodeAttribute::getExceptionTable(const u_int16_t& idx) const {
  if (idx >= exception_table.size()) {
    cerr << "Index " << idx << " for exception_table out of bounds!" << endl;
    throw exception();
  }

  return *(exception_table[idx]);
}

u_int16_t CodeAttribute::getAttributesCount() const {
  return attributes.size();
}

const AttributeInfo& CodeAttribute::getAttributeInfo(const u_int16_t& idx) const {
  if (idx >= attributes.size()) {
    cerr << "Index " << idx << " for attributes out of bounds!" << endl;
    throw exception();
  }

  return *(attributes[idx]);
}

ostream& CodeAttribute::toOStream(ostream& os) const {
  os << "Code:\t"
     << "max_stack(" << max_stack << ")\t"
     << "max_locals(" << max_locals << ")\t"
     << "code_length(" << code_length << ")\t";
  if (attributes.size() > 0) {
    os << "Attributes:\t";
    for (unsigned int i = 0; i < attributes.size(); i++) {
      os << *attributes[i] << "\t";
    }
  }
  if (exception_table.size() > 0) {
    os << "\nExceptions:\n";
    for (unsigned int i = 0; i < exception_table.size(); i++) {
      os << "from(" << exception_table[i]->getStartPC() << ") "
         << "to(" << exception_table[i]->getEndPC() << ") "
         << "handler(" << exception_table[i]->getHandlerPC() << ") ";
      if (exception_table[i]->getCatchType()) {
        os << "idx(" << exception_table[i]->getCatchType() << ")";
      } else {
        os << "<any>";
      }
      os << endl;
    }
  }

  return os;
}

ostream& CodeAttribute::toOStreamAsHTML(ostream& os) const {
  os << "<li>Code:</li>"
     << "<ul>"
     << "<li>max_stack(" << max_stack << ")</li>" << endl
     << "<li>max_locals(" << max_locals << ")</li>" << endl
     << "<li>code_length(" << code_length << ")</li>" << endl
     << "<li>Code:</li>" << endl
     << "<table bgcolor=\"#CCCCCC\"><tr><td><pre>"
     << ByteCode(code, code_length)
     << "</pre></td></tr></table>" << endl;

  if (exception_table.size() > 0) {
    os << "<li>Exceptions:</li>" << endl
       << "<table border=\"2\">" << endl
       << "<tbody>" << endl
       << "<tr>" << endl
       << "<th>StartPC</th><th>EndPC</th><th>HandlerPC</th><th>CatchType</th>" << endl
       << "</tr>";
    for (unsigned int i = 0; i < exception_table.size(); i++) {
      os << "<tr>" << endl
         << "<td>" << exception_table[i]->getStartPC() << "</td>"
         << "<td>" << exception_table[i]->getEndPC() << "</td>"
         << "<td>" << exception_table[i]->getHandlerPC() << "</td>"
         << "<td>";
      if (exception_table[i]->getCatchType()) {
        os << exception_table[i]->getCatchType();
      } else {
        os << "<i>any</i>";
      }
      os << "</td>" << endl
         << "</tr>";
    }
    os << "</tbody>" << endl
       << "</table>" << endl;
  }

  if (attributes.size() > 0) {
    os << "<li>Attributes:</li>" << endl
       << "<ul>" << endl;
    for (unsigned int i = 0; i < attributes.size(); i++) {
      attributes[i]->toOStreamAsHTML(os);
    }
    os << "</ul>" << endl;
  }
  os << "</ul>" << endl;

  return os;
}

