#include "FieldInfo.hpp"

#include "ByteClass.hpp"

#include <list>

FieldInfo::FieldInfo(const u_int16_t& af /*= 0*/,
                     const u_int16_t& ni /*= 0*/,
                     const u_int16_t& di /*= 0*/,
                     const u_int16_t& ac /*= 0*/)
  : access_flags(af),
    name_index(ni),
    descriptor_index(di),
    attributes_count(ac),
    attributes(ac) {
}

FieldInfo::~FieldInfo() {
  for (unsigned int i = 0; i < attributes.size(); i++) {
    if (attributes[i]) {
      delete attributes[i];
      attributes[i] = NULL;
    }
  }
}

const u_int16_t& FieldInfo::getAccessFlags() const {
  return access_flags;
}

const u_int16_t& FieldInfo::getNameIndex() const {
  return name_index;
}

const u_int16_t& FieldInfo::getDescriptorIndex() const {
  return descriptor_index;
}

const u_int16_t& FieldInfo::getAttributesCount() const {
  return attributes_count;
}

const AttributeInfo& FieldInfo::getAttribute(const u_int16_t& idx) const {
  if (idx >= attributes_count) {
    cerr << "Index " << idx << " for attributes out of bounds!" << endl;
    throw exception();
  }

  return *(attributes[idx]);
}

list<string> getFieldAccessFlags(const u_int16_t& access_flags) {
  list<string> access_flags_list;
  if (access_flags & ByteClass::ACC_PUBLIC) {
    access_flags_list.push_back("public");
  }
  if (access_flags & ByteClass::ACC_PRIVATE) {
    access_flags_list.push_back("private");
  }
  if (access_flags & ByteClass::ACC_PROTECTED) {
    access_flags_list.push_back("protected");
  }
  if (access_flags & ByteClass::ACC_STATIC) {
    access_flags_list.push_back("static");
  }
  if (access_flags & ByteClass::ACC_FINAL) {
    access_flags_list.push_back("final");
  }
  if (access_flags & ByteClass::ACC_VOLATILE) {
    access_flags_list.push_back("volatile");
  }
  if (access_flags & ByteClass::ACC_TRANSIENT) {
    access_flags_list.push_back("transient");
  }

  return access_flags_list;
}

ostream& FieldInfo::toOStream(ostream& os) const {
  os << "FieldInfo:"
     << "\tAccessFlags: ";
  list<string> afl(getFieldAccessFlags(getAccessFlags()));
  for (list<string>::const_iterator af = afl.begin(); af != afl.end(); af++) {
    if (af != afl.begin()) {
      os << ", ";
    }
    os << *af;
  }
  os << "\tname_idx(" << getNameIndex() << ")"
     << "\tdescriptor_idx(" << getDescriptorIndex() << ")";
  if (getAttributesCount() > 0) {
    os << "\tAttributes: ";
    for (unsigned int i = 0; i < getAttributesCount(); i++) {
      os << getAttribute(i);
      if (i+1 != getAttributesCount()) {
        os << ", ";
      }
    }
  }

  return os;
}

ostream& FieldInfo::toOStreamAsHTML(ostream& os) const {
  os << "<li>FieldInfo:</li>" << endl
     << "<ul>" << endl
     << "<li>AccessFlags: ";
  list<string> afl(getFieldAccessFlags(getAccessFlags()));
  for (list<string>::const_iterator af = afl.begin(); af != afl.end(); af++) {
    if (af != afl.begin()) {
      os << ", ";
    }
    os << *af;
  }
  os << "</li>" << endl
     << "<li>name_idx(";
  ByteClass::referenceTo(os, getNameIndex(), "ConstantPool");
  os << ")</li>" << endl
     << "<li>descriptor_idx(";
  ByteClass::referenceTo(os, getDescriptorIndex(), "ConstantPool");
  os << ")</li>" << endl;
  if (getAttributesCount() > 0) {
    os << "<li>Attributes:</li>" << endl
       << "<ul>" << endl;
    for (unsigned int i = 0; i < getAttributesCount(); i++) {
      getAttribute(i).toOStreamAsHTML(os);
    }
    os << "</ul>" << endl;
  }
  os << "</ul>" << endl;

  return os;
}

ostream& operator<<(ostream& os, const FieldInfo& fi) {
  return fi.toOStream(os);
}
